@@ -1,3 +1,23 @@
+/*
+ *  amtterm -- Intel AMT serial-over-lan client, gtk version.
+ *
+ *  Copyright (C) 2007 Gerd Hoffmann <kraxel@redhat.com
+ *
+ *  This program is free software; you can redistribute it and/or modify
+ *  it under the terms of the GNU General Public License as published by
+ *  the Free Software Foundation; either version 2 of the License, or
+ *  (at your option) any later version.
+ *
+ *  This program is distributed in the hope that it will be useful,
+ *  but WITHOUT ANY WARRANTY; without even the implied warranty of
+ *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+ *  GNU General Public License for more details.
+ *
+ *  You should have received a copy of the GNU General Public License along
+ *  with this program; if not, write to the Free Software Foundation, Inc.,
+ *  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
+ */
+
 #include <stdio.h>
 #include <stdlib.h>
 #include <unistd.h>
@@ -8,6 +28,8 @@
 #include <signal.h>
 #include <ctype.h>
 
+#include <gdk/gdk.h>
+#include <gdk/gdkx.h>
 #include <gtk/gtk.h>
 #include <vte/vte.h>
 
@@ -23,6 +45,7 @@ struct gamt_window {
     GtkWidget      *win;
     GtkWidget      *vte;
     GtkWidget      *status;
+    GtkWidget      *icon;
 
     GtkActionGroup *hosts_ag;
     guint          hosts_id;
@@ -33,11 +56,30 @@ struct gamt_window {
     guint          id;
 };
 
+static const char *state_stock[] = {
+    [ REDIR_NONE      ] = GTK_STOCK_DISCONNECT,
+#if 0
+    [ REDIR_CONNECT   ] = GTK_STOCK_,
+    [ REDIR_INIT      ] = GTK_STOCK_,
+    [ REDIR_AUTH      ] = GTK_STOCK_,
+    [ REDIR_INIT_SOL  ] = GTK_STOCK_,
+#endif
+    [ REDIR_RUN_SOL   ] = GTK_STOCK_CONNECT,
+#if 0
+    [ REDIR_INIT_IDER ] = GTK_STOCK_,
+    [ REDIR_RUN_IDER  ] = GTK_STOCK_,
+    [ REDIR_CLOSING   ] = GTK_STOCK_,
+#endif
+    [ REDIR_CLOSED    ] = GTK_STOCK_DISCONNECT,
+    [ REDIR_ERROR     ] = GTK_STOCK_DISCONNECT,
+};
+
 static char amt_host[64];
 static char amt_port[16];
 static char amt_user[32] = "admin";
 static char amt_pass[32];
 static int amt_trace;
+static int amt_debug;
 
 static int gamt_getstring(GtkWidget *window, char *title, char *message,
 			  char *dest, int dlen, int hide);
@@ -50,6 +92,7 @@ static void gamt_rebuild_hosts(struct gamt_window *gamt);
 #define CFG_FONT          CFG_SECTION, "font"
 #define CFG_FOREGROUND    CFG_SECTION, "foreground"
 #define CFG_BACKGROUND    CFG_SECTION, "background"
+#define CFG_BLINK         CFG_SECTION, "blink-cursor"
 
 /* ------------------------------------------------------------------ */
 
@@ -107,12 +150,15 @@ static void menu_cb_config_font(GtkAction *action, void *data)
     fontname = cfg_get_str(CFG_FONT);
     gtk_font_selection_dialog_set_font_name
 	(GTK_FONT_SELECTION_DIALOG(dialog), fontname);
+
     gtk_widget_show_all(dialog);
-    if (GTK_RESPONSE_OK == gtk_dialog_run(GTK_DIALOG(dialog))) {
+    switch (gtk_dialog_run(GTK_DIALOG(dialog))) {
+    case GTK_RESPONSE_OK:
 	fontname = gtk_font_selection_dialog_get_font_name
 	    (GTK_FONT_SELECTION_DIALOG(dialog));
 	vte_terminal_set_font_from_string(VTE_TERMINAL(gamt->vte), fontname);
 	cfg_set_str(CFG_FONT, fontname);
+	break;
     }
     gtk_widget_destroy(dialog);
 }
@@ -129,7 +175,8 @@ static int pickcolor(char *title, GdkColor *color)
     gtk_color_selection_set_current_color(csel, color);
     
     gtk_widget_show_all(dialog);
-    if (GTK_RESPONSE_OK == gtk_dialog_run(GTK_DIALOG(dialog))) {
+    switch (gtk_dialog_run(GTK_DIALOG(dialog))) {
+    case GTK_RESPONSE_OK:
 	gtk_color_selection_get_current_color(csel, color);
 	rc = 0;
     }
@@ -167,6 +214,17 @@ static void menu_cb_config_bg(GtkAction *action, void *data)
     cfg_set_str(CFG_BACKGROUND, name);
 }
 
+static void menu_cb_blink_cursor(GtkToggleAction *action, gpointer user_data)
+{
+    struct gamt_window *gamt = user_data;
+    gboolean state = gtk_toggle_action_get_active(action);
+
+    if (amt_debug)
+	fprintf(stderr, "%s: %s\n", __FUNCTION__, state ? "on" : "off");
+    cfg_set_bool(CFG_BLINK, state);
+    vte_terminal_set_cursor_blinks(VTE_TERMINAL(gamt->vte), state);
+}
+
 static void menu_cb_quit(GtkAction *action, void *data)
 {
     struct gamt_window *gamt = data;
@@ -174,10 +232,50 @@ static void menu_cb_quit(GtkAction *action, void *data)
     gtk_widget_destroy(gamt->win);
 }
 
+static void show_manpage(char *page, char *section)
+{
+    char buf[64];
+
+
+    switch(fork()) {
+    case -1:
+	perror("fork");
+	break;
+    case 0:
+	/* child: try xdg-open first ... */
+	snprintf(buf, sizeof(buf), "man:%s(%s)", page, section);
+	execlp("xdg-open", "xdg-open", buf, NULL);
+	perror("execlp(xdg-open)");
+	/* ... fallback is an xterm with man */
+	snprintf(buf, sizeof(buf), "manual page: %s(%s)", page, section);
+	execlp("xterm", "xterm",
+	       "-title", buf, 
+	       "-e", "man", section, page,
+	       NULL);
+	perror("execlp(xterm)");
+	exit(1);
+	break;
+    default:
+	/* parent */
+	break;
+    }
+}
+
+static void menu_cb_man_gamt(GtkAction *action, void *data)
+{
+    show_manpage("gamt", "1");
+}
+
+static void menu_cb_man_amt_howto(GtkAction *action, void *data)
+{
+    show_manpage("amt-howto", "7");
+}
+
 static void menu_cb_about(GtkAction *action, void *data)
 {
     static char *comments = "Intel AMT serial-over-lan client";
     static char *copyright = "(c) 2007 Gerd Hoffmann";
+    static char *website = "http://dl.bytesex.org/releases/amtterm/";
     static char *authors[] = { "Gerd Hoffmann <kraxel@redhat.com>", NULL };
     struct gamt_window *gamt = data;
 
@@ -187,6 +285,8 @@ static void menu_cb_about(GtkAction *action, void *data)
                           "copyright",       copyright,
                           "logo-icon-name",  GTK_STOCK_ABOUT,
                           "version",         VERSION,
+			  "website",         website,
+//			  "license",         "GPLv2+",
                           NULL);
 }
 
@@ -215,8 +315,13 @@ static void state_gtk(void *cb_data, enum redir_state old, enum redir_state new)
 
     switch (new) {
     case REDIR_ERROR:
-	snprintf(buf, sizeof(buf), "%s: %s FAILED", gamt->redir.host,
-		 redir_state_desc(old));
+#if 0
+	snprintf(buf, sizeof(buf), "%s: %s FAILED (%s)", gamt->redir.host,
+		 redir_state_desc(old), gamt->redir.err);
+#else
+	snprintf(buf, sizeof(buf), "%s: ERROR: %s", gamt->redir.host,
+		 gamt->redir.err);
+#endif
 	if (old == REDIR_AUTH) {
 	    /* ask for a new password next time ... */
 	    strcpy(amt_pass, "");
@@ -233,6 +338,9 @@ static void state_gtk(void *cb_data, enum redir_state old, enum redir_state new)
 		 redir_state_desc(new));
 	break;
     }
+    if (state_stock[new])
+	gtk_image_set_from_stock(GTK_IMAGE(gamt->icon), state_stock[new],
+				 GTK_ICON_SIZE_SMALL_TOOLBAR);
     gtk_label_set_text(GTK_LABEL(gamt->status), buf);
 }
 
@@ -286,6 +394,7 @@ static const GtkActionEntry entries[] = {
 	.callback    = G_CALLBACK(menu_cb_config_font),
     },{
 	.name        = "VteForeground",
+//	.stock_id    = GTK_STOCK_SELECT_COLOR,
 	.label       = "_Text Color ...",
 	.callback    = G_CALLBACK(menu_cb_config_fg),
     },{
@@ -295,6 +404,16 @@ static const GtkActionEntry entries[] = {
     },{
 
 	/* Help menu */
+	.name        = "ManGamt1",
+	.stock_id    = GTK_STOCK_HELP,
+	.label       = "Manual Page",
+	.callback    = G_CALLBACK(menu_cb_man_gamt),
+    },{
+	.name        = "ManAmtHowto7",
+	.stock_id    = GTK_STOCK_INFO,
+	.label       = "AMT HowTo",
+	.callback    = G_CALLBACK(menu_cb_man_amt_howto),
+    },{
 	.name        = "About",
 	.stock_id    = GTK_STOCK_ABOUT,
 	.label       = "_About ...",
@@ -302,6 +421,14 @@ static const GtkActionEntry entries[] = {
     }
 };
 
+static const GtkToggleActionEntry tentries[] = {
+    {
+	.name        = "BlinkCursor",
+	.label       = "Blinking cursor",
+	.callback    = G_CALLBACK(menu_cb_blink_cursor),
+    }
+};
+
 static char ui_xml[] =
 "<ui>\n"
 "  <menubar action='MainMenu'>\n"
@@ -317,8 +444,13 @@ static char ui_xml[] =
 "      <menuitem action='VteFont'/>\n"
 "      <menuitem action='VteForeground'/>\n"
 "      <menuitem action='VteBackground'/>\n"
+"      <separator/>\n"
+"      <menuitem action='BlinkCursor'/>\n"
 "    </menu>\n"
 "    <menu action='HelpMenu'>\n"
+"      <menuitem action='ManGamt1'/>\n"
+"      <menuitem action='ManAmtHowto7'/>\n"
+"      <separator/>\n"
 "      <menuitem action='About'/>\n"
 "    </menu>\n"
 "  </menubar>\n"
@@ -488,6 +620,7 @@ static int gamt_connect(struct gamt_window *gamt)
     if (-1 == redir_connect(&gamt->redir))
 	return -1;
 
+    fcntl(gamt->redir.sock, F_SETFD, FD_CLOEXEC);
     vte_terminal_reset(VTE_TERMINAL(gamt->vte), TRUE, TRUE);
     gamt->ch = g_io_channel_unix_new(gamt->redir.sock);
     gamt->id = g_io_add_watch(gamt->ch, G_IO_IN, gamt_data, gamt);
@@ -497,9 +630,10 @@ static int gamt_connect(struct gamt_window *gamt)
 
 static struct gamt_window *gamt_window()
 {
-    GtkWidget *vbox, *frame, *item;
+    GtkWidget *vbox, *hbox, *frame, *item;
     GdkColor color;
     GError *err;
+    gboolean state;
     struct gamt_window *gamt;
     char *str;
     
@@ -517,6 +651,8 @@ static struct gamt_window *gamt_window()
     gamt->ui = gtk_ui_manager_new();
     gamt->ag = gtk_action_group_new("MenuActions");
     gtk_action_group_add_actions(gamt->ag, entries, G_N_ELEMENTS(entries), gamt);
+    gtk_action_group_add_toggle_actions(gamt->ag, tentries,
+					G_N_ELEMENTS(tentries), gamt);
     gtk_ui_manager_insert_action_group(gamt->ui, gamt->ag, 0);
 #if 0
     GtkAccelGroup *accel = gtk_ui_manager_get_accel_group(gamt->ui);
@@ -536,16 +672,29 @@ static struct gamt_window *gamt_window()
     g_signal_connect(gamt->vte, "commit", G_CALLBACK(user_input), gamt);
     vte_terminal_set_scrollback_lines(VTE_TERMINAL(gamt->vte), 4096);
     str = cfg_get_str(CFG_FONT);
-    if (str)
-	vte_terminal_set_font_from_string(VTE_TERMINAL(gamt->vte), str);
+    vte_terminal_set_font_from_string(VTE_TERMINAL(gamt->vte), str);
 
+    /* FIXME: make configurable */
+    vte_terminal_set_backspace_binding(VTE_TERMINAL(gamt->vte),
+				       VTE_ERASE_ASCII_BACKSPACE);
+    vte_terminal_set_delete_binding(VTE_TERMINAL(gamt->vte),
+				    VTE_ERASE_AUTO);
+
+    item = gtk_ui_manager_get_widget(gamt->ui, "/MainMenu/ConfigMenu/BlinkCursor");
+    state = cfg_get_bool(CFG_BLINK, 0);
+    gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(item), state);
+    vte_terminal_set_cursor_blinks(VTE_TERMINAL(gamt->vte), state);
+    
     /* other widgets */
     gamt->status = gtk_label_new("idle");
     gtk_misc_set_alignment(GTK_MISC(gamt->status), 0, 0.5);
     gtk_misc_set_padding(GTK_MISC(gamt->status), 3, 1);
-
+    gamt->icon = gtk_image_new_from_stock(GTK_STOCK_DISCONNECT,
+					  GTK_ICON_SIZE_SMALL_TOOLBAR);
+    
     /* Make a vbox and put stuff in */
     vbox = gtk_vbox_new(FALSE, 1);
+    hbox = gtk_hbox_new(FALSE, 1);
     gtk_container_set_border_width(GTK_CONTAINER(vbox), 1);
     gtk_container_add(GTK_CONTAINER(gamt->win), vbox);
     item = gtk_ui_manager_get_widget(gamt->ui, "/MainMenu");
@@ -555,10 +704,15 @@ static struct gamt_window *gamt_window()
     gtk_box_pack_start(GTK_BOX(vbox), item, FALSE, FALSE, 0);
 #endif
     gtk_box_pack_start(GTK_BOX(vbox), gamt->vte, TRUE, TRUE, 0);
+    gtk_box_pack_end(GTK_BOX(vbox), hbox, FALSE, TRUE, 0);
 
     frame = gtk_frame_new(NULL);
-    gtk_box_pack_end(GTK_BOX(vbox), frame, FALSE, TRUE, 0);
     gtk_container_add(GTK_CONTAINER(frame), gamt->status);
+    gtk_box_pack_start(GTK_BOX(hbox), frame, TRUE, TRUE, 0);
+
+    frame = gtk_frame_new(NULL);
+    gtk_container_add(GTK_CONTAINER(frame), gamt->icon);
+    gtk_box_pack_end(GTK_BOX(hbox), frame, FALSE, TRUE, 0);
 
     /* display window */
     gtk_widget_show_all(gamt->win);
@@ -606,10 +760,10 @@ static void usage(FILE *fp)
 int
 main(int argc, char *argv[])
 {
+    Display *dpy;
     struct gamt_window *gamt;
     char configfile[256];
     char *h;
-    int debug = 0;
     int c;
 
     if (NULL != (h = getenv("AMT_PASSWORD")))
@@ -626,12 +780,15 @@ main(int argc, char *argv[])
 	cfg_set_str(CFG_BACKGROUND, "black");
 
     gtk_init(&argc, &argv);
+    dpy = gdk_x11_display_get_xdisplay(gdk_display_get_default());
+    fcntl(ConnectionNumber(dpy),F_SETFD,FD_CLOEXEC);
+
     for (;;) {
         if (-1 == (c = getopt(argc, argv, "hdtu:p:f:c:b:")))
             break;
         switch (c) {
 	case 'd':
-	    debug++;
+	    amt_debug++;
 	    break;
 	case 't':
 	    amt_trace++;
